#ifndef SYSFEATURES_INTEL_THERMAL_H
#define SYSFEATURES_INTEL_THERMAL_H

#include <sysFeatures_types.h>

extern const _SysFeatureList likwid_sysft_intel_cpu_thermal_feature_list;

#endif //SYSFEATURES_INTEL_THERMAL_H
