#ifndef SYSFEATURES_INTEL_FREQ_H
#define SYSFEATURES_INTEL_FREQ_H

#include <sysFeatures_types.h>

extern const _SysFeatureList likwid_sysft_x86_cpu_freq_feature_list;

#endif //SYSFEATURES_INTEL_FREQ_H
